import torch
from .ActivationFunction import ActivationFunction


class DerivativeSiLU(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        a = torch.sigmoid(x)
        b = 1 - torch.sigmoid(x)
        c = 1 + x * b
        return a * c



import torch
from .PatchEmbedding import PatchEmbedding


class VisionTransformerEncoder(torch.torch.nn.Module):
    def __init__(
        self,
        image_size: int,
        patch_size: int,
        input_channels: int,
        embedding_dimension: int,
        depth: int,
        num_heads: int,
        mlp_ratio: int,
        dropout: float,
    ):
        super().__init__()

        self.patch_embedding = PatchEmbedding(
            image_size=image_size,
            patch_size=patch_size,
            channels=input_channels,
            embedding_dimension=embedding_dimension,
        )
        num_patches = self.patch_embedding.num_patches
        self.cls_token = torch.nn.Parameter(torch.zeros(1, 1, embedding_dimension))
        self.pos_embed = torch.nn.Parameter(
            torch.zeros(1, num_patches + 1, embedding_dimension)
        )
        encoder_layer = torch.nn.TransformerEncoderLayer(
            d_model=embedding_dimension,
            nhead=num_heads,
            dim_feedforward=int(embedding_dimension * mlp_ratio),
            dropout=dropout,
            batch_first=True,
            activation="gelu",
        )
        self.encoder = torch.nn.TransformerEncoder(encoder_layer, num_layers=depth)
        self.norm = torch.nn.LayerNorm(embedding_dimension)
        self._init()

    def _init(self):
        torch.nn.init.normal_(self.cls_token, std=1e-6)
        torch.nn.init.normal_(self.pos_embed, std=1e-6)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        B = x.size(0)
        x = self.patch_embedding(x)
        cls = self.cls_token.expand(B, -1, -1)
        x = torch.cat((cls, x), dim=1)
        x = x + self.pos_embed
        features = self.encoder(x)
        features = self.norm(features[:, 0])
        return features



import torch


class PatchEmbedding(torch.nn.Module):
    def __init__(self, patch_size: int, channels: int, embedding_dimension: int):
        super().__init__()
        self.patch_size: int = patch_size
        self.channels: int = channels
        self.phi = torch.nn.Linear(channels * patch_size**2, embedding_dimension)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        batch_size, _, _, _ = x.shape
        x = x.unfold(2, self.patch_size, self.patch_size)
        x = x.unfold(3, self.patch_size, self.patch_size)
        x = x.contiguous().view(batch_size, -1, self.channels * self.patch_size**2)
        return self.phi(x)



import torch
import random
from typing import List, Type


def replace_modules(
    module, needle: torch.nn.Module, candidates: List[Type[torch.nn.Module]]
):
    for name, child in module.named_children():
        if isinstance(child, needle):
            new_activation = random.choice(candidates)
            setattr(module, name, new_activation())
        else:
            replace_modules(child, needle, candidates)
    return module



import torch
from .ActivationFunction import ActivationFunction


class ShiftedReLU(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor):
        return x.clamp(min=-1)



class SavesEnsembleModels:
    def __init__(self):
        super().__init__()

    def save(self):
        self.save_factories()
        self.save_weights()

    def save_factories(self):
        pass

    def save_weights(self):
        pass



import torch
from .ActivationFunction import ActivationFunction


class TSiLU(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        alpha = x / (1 + torch.exp(-x))
        beta = torch.exp(alpha) - torch.exp(-alpha)
        theta = torch.exp(alpha) + torch.exp(alpha)
        return beta / theta



import torch


class SelfAttentionModule(torch.nn.Module):
    def __init__(self, dimension: int):
        super().__init__()
        self.Q = torch.nn.Linear(dimension, dimension)
        self.K = torch.nn.Linear(dimension, dimension)
        self.V = torch.nn.Linear(dimension, dimension)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        Q = self.Q(x)
        K = self.K(x)
        V = self.V(x)

        scores = Q @ K.transpose(0, 1)
        scores = scores / (x.size(-1) ** 0.5)
        scores = torch.nn.functional.softmax(scores, dim=-1)

        return scores @ V



import torch
from torch.distributions import Normal
from .modules import ContinuousActorModule
from ...common.concerns import MovesModulesToDevice


class Actor(MovesModulesToDevice):
    def __init__(
        self,
        state_dimension: int,
        action_dimension: int,
        num_skills: int,
        hidden_dimension: int,
        lr: float,
        decay: float,
        epsilon: float,
        min_std: float,
        max_std: float,
    ):
        self.epsilon: float = epsilon
        self.network = ContinuousActorModule(
            state_dimension=state_dimension,
            action_dimension=action_dimension,
            hidden_dimension=hidden_dimension,
            num_skills=num_skills,
            min_std=min_std,
            max_std=max_std,
        )
        self.optimizer = torch.optim.AdamW(
            self.network.parameters(), lr=lr, weight_decay=decay
        )
        self.move_modules()

    def action(self, state: torch.Tensor, skill: torch.Tensor):
        mean, std = self.network(state, skill)
        std = std.exp()
        distribution = Normal(mean, std)
        raw_action = distribution.rsample()
        action = torch.tanh(raw_action)
        corrected_probabilities = torch.log(1.0 - action.pow(2) + self.epsilon)
        probabilities = distribution.log_prob(raw_action) - corrected_probabilities
        probabilities = probabilities.sum(dim=-1, keepdim=True)
        return action, probabilities

    def train(self, advantages: torch.Tensor, probabilities: torch.Tensor) -> float:
        self.optimizer.zero_grad()
        loss = self.loss(advantages=advantages, probabilities=probabilities)
        loss.backward()
        self.optimizer.step()
        return loss.item()

    def loss(
        self, advantages: torch.Tensor, probabilities: torch.Tensor
    ) -> torch.Tensor:
        return -(probabilities * advantages.detach()).mean()



import torch
from .ActivationFunction import ActivationFunction


class NLReLU(ActivationFunction):
    def __init__(self, beta: float = 1.05):
        super().__init__()
        self.beta = beta

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.log(self.beta * x.clamp(min=0) + 1)



class HasRecordAlternatives:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add(self, *args, **kwargs) -> None:
        self.record(*args, **kwargs)

    def push(self, *args, **kwargs) -> None:
        self.record(*args, **kwargs)

    def append(self, *args, **kwargs) -> None:
        self.record(*args, **kwargs)



import torch
from ...AdaptiveActivationFunction import AdaptiveActivationFunction
from ...concerns import ChannelBased


class StarReLU(AdaptiveActivationFunction, ChannelBased):
    def __init__(self):
        super().__init__()
        self.a = None
        self.b = None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        self.initialize(x, ["a", "b"])
        a = self.a.view(self.parameter_shape(x))
        b = self.b.view(self.parameter_shape(x))
        return a * torch.relu(x).pow(2) + b



import torch
from .ActivationFunction import ActivationFunction


class FlattedTSwish(ActivationFunction):
    def __init__(self, t: float = -0.2):
        super().__init__()
        self.t = t

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.relu(x) * torch.sigmoid(x) + self.t



import torch
from .ActivationFunction import ActivationFunction


class mReLU(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.min(
            torch.nn.functional.relu(1 - x),
            torch.nn.functional.relu(1 + x),
        )



import torch
from .ActivationFunction import ActivationFunction


class RePU(ActivationFunction):
    def __init__(self, alpha: float = 1.0):
        super().__init__()
        self.alpha = alpha

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.relu(x.pow(self.alpha))



import torch
from .ActivationFunction import ActivationFunction


class MSiLU(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        alpha = torch.exp(-x.pow(2) - 1)
        delta = alpha / 4
        return x * torch.sigmoid(x) + delta



from collections import deque
from typing import Type
from .concerns import ResetsCollection, HasRecordAlternatives


class Buffer(ResetsCollection, HasRecordAlternatives):
    def __init__(self, capacity: int):
        self.capacity: int = capacity
        self.collection: Type[deque] = deque(maxlen=self.capacity)

    @property
    def size(self):
        return len(self.collection)



import torch
from .modules import QuantileCriticModule


class QuantileMultiCritic:
    def __init__(
        self,
        state_dimension: int,
        action_dimension: int,
        hidden_dimension: int,
        num_critics: int,
        num_quantiles: int,
    ):
        super().__init__()
        self.networks = torch.nn.ModuleList(
            [
                QuantileCriticModule(
                    state_dimension=state_dimension,
                    action_dimension=action_dimension,
                    hidden_dimension=hidden_dimension,
                    num_quantiles=num_quantiles,
                )
                for _ in range(num_critics)
            ]
        )
        self.targets = torch.nn.ModuleList(
            [
                QuantileCriticModule(
                    state_dimension=state_dimension,
                    action_dimension=action_dimension,
                    hidden_dimension=hidden_dimension,
                    num_quantiles=num_quantiles,
                )
                for _ in range(num_critics)
            ]
        )
        for network, target in zip(self.networks, self.targets):
            target.load_state_dict(network.state_dict())



import torch
from .Loss import Loss


class LogitNormLoss(Loss):
    def __init__(self, epsilon: float = 1e-7):
        super().__init__()
        self.epsilon = epsilon

    def forward(self, logits: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        norms = torch.norm(logits, p=2, dim=-1, keepdim=True)
        norms = norms + self.epsilon
        normalized_logits = torch.div(logits, norms)
        return torch.nn.functional.cross_entropy(normalized_logits, target)



import torch
from .NoisyLinear import NoisyLinear


class NoisyFeedForward(torch.nn.Module):
    def __init__(
        self,
        input_dimension: int,
        hidden_dimension: int,
        output_dimension: int,
        activation_fn=torch.nn.GELU,
    ):
        super().__init__()
        self.mu = NoisyLinear(input_dimension, hidden_dimension)
        self.xi = NoisyLinear(hidden_dimension, output_dimension)
        self.chi = activation_fn()

    def reset_noise(self):
        self.mu.reset_noise()
        self.xi.reset_noise()

    def forward(self, x):
        x = self.mu(x)
        x = self.chi(x)
        x = self.xi(x)
        return x



import torch
from .Classifier import Classifier
from .concerns import Trainable, ReportsMetrics
from .modules import SwinModule
from ..common.concerns import MovesModulesToDevice


class Swin(Classifier, Trainable, ReportsMetrics, MovesModulesToDevice):
    def __init__(
        self,
        num_classes: int,
        lr: float = 0.0001,
        decay: float = 0.0,
        variant: str = "base",
        weights="DEFAULT",
    ):
        super().__init__()

        self.num_classes: int = num_classes

        self.network = SwinModule(
            num_classes=self.num_classes, variant=variant, weights=weights
        )
        self.move_modules()

        self.optimizer = torch.optim.AdamW(
            self.network.parameters(), lr=lr, weight_decay=decay
        )
        self.criterion = torch.nn.CrossEntropyLoss()



import torch
from .ActivationFunction import ActivationFunction


class IpLU(ActivationFunction):
    def __init__(self, alpha: float = 1.4):
        super().__init__()
        self.alpha = alpha

    def forward(self, x: torch.Tensor):
        delta = 1 + x.abs().pow(self.alpha)
        return torch.where(x >= 0, x, x / delta)



import torch
from .ActivationFunction import ActivationFunction


class SigLU(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        delta = torch.exp(-2 * x)
        return torch.where(x >= 0, x, (1 - delta) / (1 + delta))



from .sac import SoftActor, SoftCritic, SoftActorCriticTemperature
from .buffers import SoftActorCriticBuffer
from ..common.concerns import TracksTime
from ..device import get_device


class SoftActorCritic(TracksTime):
    def __init__(
        self,
        state_dimension: int,
        action_dimension: int,
        hidden_dimension: int = 256,
        buffer_capacity: int = 500_000,
        batch_size: int = 256,
        actor_lr: float = 0.0003,
        critic_lr: float = 0.0003,
        temperature_lr: float = 0.0003,
        actor_decay: float = 0.0,
        critic_decay: float = 0.0,
        temperature_decay: float = 0,
        tau: float = 0.005,
        gamma: float = 0.99,
        min_std: float = -20.0,
        max_std: float = 2.0,
        warmup: int = 20_000,
        device: None | str = None,
    ):
        super().__init__()

        self.batch_size: int = batch_size
        self.tau: float = tau
        self.warmup: int = warmup

        if device is None:
            device = get_device()

        self.actor = SoftActor(
            state_dimension=state_dimension,
            action_dimension=action_dimension,
            hidden_dimension=hidden_dimension,
            lr=actor_lr,
            weight_decay=actor_decay,
            min_std=min_std,
            max_std=max_std,
        )
        self.critic = SoftCritic(
            state_dimension=state_dimension,
            action_dimension=action_dimension,
            hidden_dimension=hidden_dimension,
            lr=critic_lr,
            weight_decay=critic_decay,
            gamma=gamma,
        )
        self.temperature = SoftActorCriticTemperature(
            lr=temperature_lr,
            weight_decay=temperature_decay,
            action_dimension=action_dimension,
            device=device,
        )
        self.buffer = SoftActorCriticBuffer(capacity=buffer_capacity)

    def is_warmup(self):
        return self.t < self.warmup

    def train(self):
        # don't train before warmup
        if self.is_warmup():
            return

        data = self.buffer.sample_torch(self.batch_size)
        alpha = self.temperature.alpha
        self.critic.train(
            states=data.states,
            actions=data.actions,
            rewards=data.rewards,
            terminations=data.terminations,
            next_states=data.next_states,
            actor=self.actor,
            alpha=alpha,
        )
        self.critic.update(tau=self.tau)
        self.actor.train(states=data.states, critic=self.critic, alpha=alpha)
        self.temperature.train()



import torch
from .concerns import MovesModulesToDevice
from .modules import EncoderModule


class FeedForward(MovesModulesToDevice):
    def __init__(
        self,
        lr: float,
        weight_decay: float,
        hidden_dimension: int,
        input_dimension: int,
    ):
        self.network = EncoderModule(
            input_dimension=input_dimension, hidden_dimension=hidden_dimension
        )
        self.optimizer = torch.optim.AdamW(
            self.network.parameters(), lr=lr, weight_decay=weight_decay
        )
        self.move_modules()



import torch
import math


class NoisyLinear(torch.nn.Module):
    def __init__(self, in_features, out_features, std_init=0.017):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.std_init = std_init

        self.weight_mu = torch.nn.Parameter(torch.empty(out_features, in_features))
        self.weight_sigma = torch.nn.Parameter(torch.empty(out_features, in_features))
        self.bias_mu = torch.nn.Parameter(torch.empty(out_features))
        self.bias_sigma = torch.nn.Parameter(torch.empty(out_features))

        self.register_buffer("eps_in", torch.empty(in_features))
        self.register_buffer("eps_out", torch.empty(out_features))

        self.reset_parameters()
        self.reset_noise()

    def reset_parameters(self):
        mu_range = 1 / math.sqrt(self.in_features)
        self.weight_mu.data.uniform_(-mu_range, mu_range)
        self.bias_mu.data.uniform_(-mu_range, mu_range)

        sigma_weight = self.std_init / math.sqrt(self.in_features)
        sigma_bias = self.std_init / math.sqrt(self.out_features)

        self.weight_sigma.data.fill_(sigma_weight)
        self.bias_sigma.data.fill_(sigma_bias)

    @staticmethod
    def f(x):
        return torch.sign(x) * torch.sqrt(torch.abs(x))

    def reset_noise(self):
        eps_in = torch.randn(self.in_features)
        eps_out = torch.randn(self.out_features)

        self.eps_in.copy_(self.f(eps_in))
        self.eps_out.copy_(self.f(eps_out))

    def forward(self, x):
        if self.training:
            noise_w = torch.outer(self.eps_out, self.eps_in)
            w = self.weight_mu + self.weight_sigma * noise_w
            b = self.bias_mu + self.bias_sigma * self.eps_out
        else:
            w = self.weight_mu
            b = self.bias_mu
        return torch.nn.functional.linear(x, w, b)



import torch


class DQNModule(torch.nn.Module):
    def __init__(self, action_dim: int, input_channels: int, embedding_dimension: int):
        super().__init__()
        self.phi = torch.nn.Sequential(
            torch.nn.Conv2d(input_channels, 32, kernel_size=8, stride=4),
            torch.nn.Conv2d(32, 64, kernel_size=4, stride=2),
            torch.nn.Conv2d(64, 64, kernel_size=3, stride=1),
        )
        self.sigma = torch.nn.Sequential(
            torch.nn.LazyLinear(embedding_dimension),
            torch.nn.ReLU(),
            torch.nn.Linear(embedding_dimension, action_dim),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        features = self.phi(x)
        return self.sigma(features)



from .ReportsEnsembleF1 import ReportsEnsembleF1
from .ReportsEnsembleAccuracy import ReportsEnsembleAccuracy
from .ReportsEnsembleKappa import ReportsEnsembleKappa
from .CalculatesMetricNecessities import CalculatesMetricNecessities


class ReportsClassificationMetrics(
    CalculatesMetricNecessities,
    ReportsEnsembleAccuracy,
    ReportsEnsembleF1,
    ReportsEnsembleKappa,
):
    def __init__(self):
        super().__init__()



import torch
import torch.nn as nn
import torch.nn.functional as F


class GraphAttentionHeadModule(nn.Module):
    def __init__(
        self,
        input_dimension: int,
        output_dimension: int,
        dropout: float,
        alpha: float,
        activation: nn.Module,
        final_activation: nn.Module,
        v2: bool,
        use_layernorm: bool,
    ):
        super().__init__()
        self.v2: bool = v2

        self.phi = nn.Linear(input_dimension, output_dimension, bias=True)
        self.mu = nn.Linear(output_dimension, 1, bias=False)
        self.xi = nn.Linear(output_dimension, 1, bias=False)

        if activation is nn.LeakyReLU:
            self.att_activation = nn.LeakyReLU(alpha)
        else:
            self.att_activation = nn.LeakyReLU(alpha)  # for e_{ij}

        self.final_activation = final_activation()
        self.dropout = nn.Dropout(dropout)
        self.norm = nn.LayerNorm(output_dimension) if use_layernorm else nn.Identity()

    def forward(self, features: torch.Tensor, adj: torch.Tensor):
        h = self.phi(features)  # [N, out]
        h = self.norm(h)

        if not self.v2:
            e = self.mu(h) + self.xi(h).T
        else:
            scores = h.unsqueeze(1) + h.unsqueeze(0)  # [N, N, out]
            e = self.mu(scores).squeeze(-1)  # [N, N]

        e = self.att_activation(e)
        e = e.masked_fill(adj == 0, float("-inf"))

        a = F.softmax(e, dim=1)
        a = self.dropout(a)

        h_out = a @ h  # [N, out]

        return self.final_activation(h_out)



import torch
from .ActivationFunction import ActivationFunction


class DiffELU(ActivationFunction):
    def __init__(self):
        super().__init__()
        self.a = 0.3
        self.b = 0.1

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        delta = x * x.exp() - self.b * torch.exp(self.b * x)
        return torch.where(x >= 0, x, self.a * delta)



import torch


@torch.no_grad()
def soft_update(network: torch.nn.Module, target: torch.nn.Module, tau: float):
    for n, t in zip(network.parameters(), target.parameters()):
        t.data.copy_(tau * n.data + (1 - tau) * t.data)



import torch
from .ActivationFunction import ActivationFunction


class EANAF(ActivationFunction):
    def __init__(self):
        super().__init__()

    def g(self, x: torch.Tensor) -> torch.Tensor:
        return torch.tanh(x / 2)

    def h(self, x: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.softplus(x)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * self.g(self.h(x))



import torch


class CriticModule(torch.nn.Module):
    def __init__(
        self,
        state_dimension: int,
        hidden_dimension: int,
        num_skills: int,
    ):
        super().__init__()
        self.state_dimension: int = state_dimension
        self.num_skills: int = num_skills
        self.hidden_dimension: int = hidden_dimension
        self.input_dimension: int = self.state_dimension + self.num_skills

        self.phi = torch.nn.Sequential(
            torch.nn.Linear(self.input_dimension, self.hidden_dimension),
            torch.nn.ReLU(),
            torch.nn.Linear(self.hidden_dimension, self.hidden_dimension),
            torch.nn.ReLU(),
        )
        self.fc = (torch.nn.Linear(self.hidden_dimension, 1),)

    def forward(self, state: torch.Tensor, skill: torch.Tensor) -> torch.Tensor:
        psi = torch.cat([state, skill], dim=-1)
        features = self.phi(psi)
        return self.fc(features).squeeze(-1)



import torch
from .ActivationFunction import ActivationFunction


class LogSigmoid(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor):
        return torch.log(torch.sigmoid(x))



import torch
import torch.nn as nn

from .GraphAttentionHeadModule import GraphAttentionHeadModule


class MultiHeadGraphAttentionModule(nn.Module):
    def __init__(
        self,
        input_dimension: int,
        output_dimension: int,
        num_heads: int,
        dropout: float,
        alpha: float,
        concat: bool,
        activation: nn.Module,
        final_activation: nn.Module,
        v2: bool,
        use_layernorm: bool,
    ):
        super().__init__()

        self.concat = concat

        self.heads = nn.ModuleList(
            [
                GraphAttentionHeadModule(
                    input_dimension=input_dimension,
                    output_dimension=output_dimension,
                    dropout=dropout,
                    alpha=alpha,
                    activation=activation,
                    final_activation=final_activation,
                    v2=v2,
                    use_layernorm=use_layernorm,
                )
                for _ in range(num_heads)
            ]
        )

    def forward(self, features: torch.Tensor, adjacency_matrix: torch.Tensor):
        outputs = [head(features, adjacency_matrix) for head in self.heads]

        if self.concat:
            return torch.cat(outputs, dim=1)
        else:
            return torch.mean(torch.stack(outputs, dim=0), dim=0)



import torch
from typing import Type


class FeedForward(torch.nn.Module):
    def __init__(
        self,
        dimension: int,
        factor: int | None = None,
        output_dimension: int | None = None,
        embedding_dimension: int | None = None,
        activation: Type[torch.nn.Module] = torch.nn.GELU,
        use_skip: bool = False,
        use_layernorm: bool = False,
    ):
        super().__init__()

        if factor is not None and embedding_dimension is not None:
            raise Exception(
                "You should specify either factor or embedding dimensions in homa.nn.FeedForward. Got both."
            )

        if factor is not None and embedding_dimension is None:
            sigma = factor * dimension
        else:
            sigma = embedding_dimension

        self.use_skip = use_skip
        self.use_layernorm = use_layernorm

        if not output_dimension:
            output_dimension = dimension

        if use_layernorm:
            self.chi = torch.nn.LayerNorm(dimension)

        self.mu = torch.nn.Linear(dimension, sigma)
        self.xi = torch.nn.Linear(sigma, output_dimension)
        self.sigma = activation()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        residual = x
        x = self.mu(x)
        x = self.sigma(x)
        x = self.xi(x)

        if self.use_skip:
            x += residual

        if self.use_layernorm:
            x = self.chi(x)

        return x



import torch
from .modules import DQNModule


class DQNAgent:
    def __init__(
        self,
        action_dim: int,
        embedding_dimension: int,
        input_channels: int,
        lr: float,
        weight_decay: float,
    ):
        self.network = DQNModule(
            action_dim=action_dim,
            embedding_dimension=embedding_dimension,
            input_channels=input_channels,
        )
        self.optimizer = torch.nn.AdamW(
            self.network.parameters(), lr=lr, weight_decay=weight_decay
        )
        self.criterion = torch.nn.SmoothL1Loss()

    def train(self):
        self.optimizer.zero_grad()
        loss = self.loss()
        loss.backward()
        self.optimizer.step()

    def loss(self):
        pass



from sklearn.metrics import f1_score as f1


class ReportsEnsembleF1:
    def __init__(self):
        super().__init__()

    def f1(self) -> float:
        predictions, labels = self.metric_necessities()
        return f1(labels, predictions, average="weighted")



import torch
from .ActivationFunction import ActivationFunction


class OAF(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.relu(x) + x * torch.sigmoid(x)



import torch
from .ReportsLogits import ReportsLogits


class PredictsProbabilities(ReportsLogits):
    def __init__(self):
        super().__init__()

    def predict(self, x: torch.Tensor) -> torch.Tensor:
        logits = self.logits(x)
        return torch.nn.functional.softmax(logits, dim=1)

    @torch.no_grad()
    def predict_(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict(x)



from sklearn.metrics import accuracy_score as accuracy
from torch.utils.data import DataLoader


class ReportsEnsembleAccuracy:
    def __init__(self):
        super().__init__()

    def accuracy(self, dataloader: DataLoader) -> float:
        predictions, labels = self.metric_necessities(dataloader)
        return accuracy(labels, predictions)



import torch
from .ActivationFunction import ActivationFunction


class PLAF(ActivationFunction):
    def __init__(self, d: float = 2):
        super().__init__()
        self.d = d

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        delta = 1 - (1 / self.d)
        sigma = (1 / self.d) * torch.abs(x).pow(self.d)
        return torch.where(x >= 1, x - delta, torch.where(x < -1, -x - delta, sigma))



import torch
from typing import Type


class EncoderModule(torch.nn.Module):
    def __init__(
        self,
        input_dimension: int,
        hidden_dimension: int,
        activation: Type[torch.nn.Module] = torch.nn.GELU,
    ):
        super().__init__()

        self.stream = torch.nn.Sequential(
            torch.nn.Linear(input_dimension, hidden_dimension),
            torch.nn.LayerNorm(hidden_dimension),
            activation(),
            torch.nn.Linear(hidden_dimension, hidden_dimension),
            torch.nn.LayerNorm(hidden_dimension),
            activation(),
        )

    def forward(self, x):
        return self.stream(x)



import torch
from ...concerns import ChannelBased
from ...AdaptiveActivationFunction import AdaptiveActivationFunction


class AOAF(AdaptiveActivationFunction, ChannelBased):
    def __init__(self, b: float = 0.17, c: float = 0.17):
        super().__init__()
        self.a = None
        self.b = b
        self.c = c

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        self.initialize(x, "a")
        a = self.a.view(self.parameter_shape(x))
        return torch.relu(x - self.b * a) + self.c * a



import torch
from ...device import get_device


class CalculatesMetricNecessities:
    def __init__(self):
        super().__init__()

    @torch.no_grad()
    def metric_necessities(self, dataloader):
        predictions, labels = [], []
        device = get_device()
        for x, y in dataloader:
            x, y = x.to(device), y.to(device)
            sum_logits = None
            for model in self.models:
                model.to(device)
                model.eval()
                logits = model(x)
                sum_logits = logits if sum_logits is None else sum_logits + logits
            batch_predictions = sum_logits.argmax(dim=1)
            predictions.extend(batch_predictions.cpu().numpy())
            labels.extend(y.cpu().numpy())
        return predictions, labels



import torch
from torchvision.models import swin_v2_b, swin_v2_s, swin_v2_t
from torch.nn.init import kaiming_uniform_ as kaiming


class SwinModule(torch.nn.Module):
    def __init__(self, num_classes: int, variant: str, weights):
        super().__init__()
        self._create_encoder(variant=variant, weights=weights)
        self._create_fc(num_classes=num_classes)

    def variant_instance(self, variant: str):
        variant_map = {"tiny": swin_v2_t, "small": swin_v2_s, "base": swin_v2_b}
        return variant_map.get(variant)

    def _create_encoder(self, variant: str, weights):
        if variant not in ["tiny", "small", "base"]:
            raise ValueError(
                f"Swin variant needs to be one of [tiny, small, base]. Invalid {variant} was provided."
            )
        instance = self.variant_instnace(variant)
        self.encoder = instance(weights=weights)
        self.encoder.head = torch.nn.Identity()

    def _create_fc(self, num_classes: int):
        self.fc = torch.nn.Linear(1024, num_classes)
        kaiming(self.fc.weight, mode="fan_in", nonlinearity="relu")

    def forward(self, images: torch.Tensor):
        features = self.encoder(images)
        return self.fc(features)
